import logging
import numpy as np

from matchernet.log import logging_conf
from matchernet.matchernet import Bundle, Matcher
from matchernet import state
from matchernet import utils

logging_conf.set_logger_config("./log/logging.json")
logger = logging.getLogger(__name__)


class BundleEKFContinuousTime(Bundle):
    """Class BundleEKFContinuousTime is a Bundle part of an extended Kalman filter (EKF) model implemented as the BundleNet.

    See matchernet.py for general concept of the BundleNet, especially the general relationship between the "Bundle" and the "Matcher".

    This class "BundleEKFContinuousTime" manages an internal state  x  and its dynamics model  f.
    The dynamics is defined as a probabilistic differential equation,
       dx/dt = f(x) + Q dw,
    where  x  is an  n  dimensional state vector variable,
           f  determines the intrinsic non-linear dynamics
                 by a map from n-dim to n-dim,
           dw is a Wiener process.
    This dynamics is locally linearlized and temporarily descritized
    as followings,
       x(t+dt) = dot( x(t), F ) + w(t), w(t) ~ N(0, dt*Q )
    where  F = exp( dt*A )  is the linear coefficient matrix
    that is defined with the Jacobian matrix  A = df/dx.

    This class stores the internal state as a temporal estimation of
    the posterior probability density function
       q(x) = N( mu, Sigma )
    that is parameterized as
           mu = self.state["mu"]  (1,n)-np.array
    and
           Sigma = self.state["Sigma"]  (n,n)-np.array.

    Dynamics function  f  is stored as a function instance
        self.f
    of a class  Fn   which is defined in  fn.py .


    In EKF algorithm, the BundleEKFContinuousTime updates
       posterior  q(x)  through following steps.

       1. Updates posterior  q(x)  according to the intrinsic dynamics
       with an arbitrary time step  dt.
       It is implemented in the method  self.step_dynamics().

       2. Updates posterior  q(x)  in order to increase objective functions
       that are defined in Matchers.
       It is implemented in the method  self.accept_feedback().


    One Bundle can connect to multiple Matchers. Thus, the difference signals
    from all the connected Matchers are applied simultaneously.

    The dynamics function  f  is fixed at the initialized value.
    See an extented class
        BundleEKFContinuousTimeTrainable (to be implemented soon)
    for a similar class with trainable dynamics function.

    """

    def __init__(self, name, dt, f, Q, mu, Sigma, logger=logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.state = state.StateMuSigma(mu, Sigma)
        self._initialize_control_params(dt)
        self.f = f
        self.Q = Q
        self.record = None
        super(BundleEKFContinuousTime, self).__init__(name, logger)

    def __call__(self, inputs):
        """The main routine that is called from brica.
        """
        for key in inputs:  # key is one of the matcher names
            if inputs[key] is not None:
                self.logger.debug("accepting feedback from {}".format(key))
                self.accept_feedback(inputs[key])
        self.step_dynamics(self.dt)
        self._countup()
        self._state_record()
        self.logger.debug("mu={}".format(self.state.data["mu"]))

        # Send state to Matcher
        results = {
            "mu": self.state.data["mu"],
            "Sigma": self.state.data["Sigma"],
            "time_stamp": self.time_stamp,
            "time_id": self.time_id
        }
        return {
            "state" : results
        }

    def _initialize_control_params(self, dt):
        self.time_id = 0
        self.time_stamp = 0.0
        self.dt = dt

    def _countup(self):
        self.time_id += 1
        self.time_stamp += self.dt

    def _state_record(self):
        mu = np.array(self.state.data["mu"], dtype=np.float32)
        sigma = np.array([np.diag(self.state.data["Sigma"])], dtype=np.float32)
        ts = np.array([self.time_stamp], dtype=np.float32)

        if self.record is None:
            self.record = {
                "mu": mu,
                "diagSigma": sigma,
                "time_stamp": ts
            }
        else:
            self.record["mu"] = np.vstack((self.record["mu"], mu))
            self.record["diagSigma"] = np.concatenate((self.record["diagSigma"], sigma), axis=0)
            self.record["time_stamp"] = np.concatenate((self.record["time_stamp"], ts), axis=0)

    def logger_state(self):
        """ Print the state of the current Bundle.
        Args:
            None.
        Returns:
            None.
        """
        self.logger.debug("State of {n}: {x}".format(n=self.name, x=self.state.data))

    def accept_feedback(self, fbst):
        """Overriding matchernet.Bundle.accept_feedback()
        This method updates the state of the current Bundle with accepting a feedback, fbst, from each Matcher linking from the current Bundle, according to the following update rule.

            mu <-- mu + dmu,
            Sigma <-- Sigma - dSigma,

        where  mu  and  Sigma  stands for the probabilistic state of the current bundle  q(x) = N( mu, Sigma ),
        and  dmu  and  dSigma  stands for the feedback state coming from the corresponding Matcher.
        """
        dmu = fbst["mu"]
        dSigma = fbst["Sigma"]
        mu = self.state.data["mu"]
        Sigma = self.state.data["Sigma"]

        self.logger.debug("dmu={}".format(dmu))
        self.logger.debug("dSigma={}".format(dSigma))
        
        weight = 1.0
        self.state.data["mu"] = (mu + weight * dmu).astype(np.float32)
        self.state.data["Sigma"] = (Sigma + weight * dSigma).astype(np.float32)
        # self.Q = (1-weight*self.lr) * Q + weight*self.lr * np.dot(dmu.T,dmu)

    def step_dynamics(self, dt):
        """This method updates  self.state  using the dynamics model,

            dx/dt  =  f(x) + Q dw.

        In order to (1) temporally discretization, (2) locally liniarization, and (3) consideration of probabilistic state, the actual update rule becomes the following manner

            mu <-- mu + dot( x(t), F )
            Sigma <-- F' * Sigma * F + dt * Q

        where  F = exp( dt*A )  is a linear coefficient matrix calculated with the Jacobian matrix  A = (df/dx).
        """
        mu = self.state.data["mu"]
        Sigma = self.state.data["Sigma"]
        A = self.f.x(mu)
        # Note:  mu.shape = (n, ), A.shape = (n,n)
        F = utils.calc_matrix_F(A, dt)
        mu = F @ mu
        Sigma = dt * self.Q + F.T @ Sigma @ F
        Sigma = utils.regularize_cov_matrix(Sigma)
        self.state.data["mu"] = mu
        self.state.data["Sigma"] = Sigma
        # ["time_stamp"] is updated in the method self._countup()


class BundleEKFWithController(Bundle):
    """Class BundleEKFWithController is a extentin of BundleEKFContinuousTime class with controller input  u.
    """
    
    def __init__(self, name, dt, f, Q, mu, Sigma, control_src_name):
        self.f = f
        self._initialize_control_params(dt)
        self._initialize_state(mu, Sigma)
        
        self.Q = Q
        self.control_src_name = control_src_name

        super(BundleEKFWithController, self).__init__(name)

    def __call__(self, inputs):
        # Update state
        for key in inputs:  # key is one of the matcher names
            if inputs[key] is not None:
                if key != self.control_src_name:
                    self.accept_feedback(inputs[key])

        # Step dynamics with applied control signal
        for key in inputs:  # key is one of the matcher names
            if inputs[key] is not None:
                if key == self.control_src_name:
                    u = inputs[key]["u"]
                    self.step_dynamics(u, self.dt)

        # Update timestamp
        self._countup()

        # Send state to Matcher
        results = {
            "mu": self.state.data["mu"],
            "Sigma": self.state.data["Sigma"],
            "time_stamp": self.time_stamp,
            "time_id": self.time_id
        }
        return {
            "state" : results
        }

    def _initialize_control_params(self, dt):
        self.dt = dt
        self.time_stamp = 0.0
        self.time_id = 0

    def _countup(self):
        self.time_stamp += self.dt
        self.time_id += 1

    def _initialize_state(self, mu, Sigma):
        self.state = state.StateMuSigma(mu, Sigma)

    def accept_feedback(self, fbst):
        dmu = fbst["mu"]
        dSigma = fbst["Sigma"]
        mu = self.state.data["mu"]
        Sigma = self.state.data["Sigma"]

        self.state.data["mu"] = (mu + dmu).astype(np.float32)
        self.state.data["Sigma"] = (Sigma + dSigma).astype(np.float32)

    def step_dynamics(self, u, dt):
        mu = self.state.data["mu"]
        Sigma = self.state.data["Sigma"]
        A = self.f.x(mu, u)
        F = utils.calc_matrix_F(A, dt)
        
        mu = mu + self.f.value(mu, u) * dt
        Sigma = dt * self.Q + F.T @ Sigma @ F
        Sigma = utils.regularize_cov_matrix(Sigma)

        self.state.data["mu"] = mu
        self.state.data["Sigma"] = Sigma
        # ["time_stamp"] is updated in the method self._countup()


class MatcherEKF(Matcher):
    """Class MatcherEKF is a Matcher part of an extended Kalman filter (EKF) model implemented as the BundleNet.

    See matchernet.py for general concept of the BundleNet, especially the general relationship between the "Bundle" and the "Matcher".

    In short, the followings are what MatcherEKF does in the BundleNet.
     (1) Defines a link between two bundles  b0  and  b1
     (2) Receives current states of bundles  b0  and  b1, which are probability density function  q(x0) = N( mu0, Sigma0 )  and  q(x1) = N( mu1, Sigma1 )
     (3) Evaluates error  E  (or log likelihood  lnL  )  defined as a function of the current bundle states.
     (4) Calculates the derivatives of  E  w.r.t. the states,
         dmu0 = d E / d mu0, dSigma0 = d E / d Sigma0
         dmu1 = d E / d mu1, dSigma1 = d E / d Sigma1
     (5) Returns the derivatives as "feedback state" from the current Matcher to the corresponding Bundles.

    The MatcherEKF defines a error (or contradiction) between the two bundles as

       E =  z S^{-1} z'

    and corresponding log-likelihood function as

       lnl = - (1/2) z S^{-1} z' - (1/2) ln det 2 pi S

    where

       z = g0( mu0 ) - g1( mu1 )

    stands for the error (or contradiction) between the states of the two Bundles.  g0()  and  g1()  are arbitrary user-defined functions. The user can set  g0  and  g1  identity function  mu0 = g0(mu0)  and  mu1 = g1(mu1)  for the simplest case.
    State variables of the bundles are denoted as normal distributions

       q(x0)  ==  N( mu0,  Sigma0 )
       q(x1)  ==  N( mu1,  Sigma1 )

    Then, the total variance  S  of the current matcher is defined as

       S = C0 Sigma0 C0' + C1 Sigma1 C1'

    where  C0 = (dg0/dx)  and  C1 = (dg1/dx)  are Jacobian matrices. Note that C0 and C1 are identity matrices and  S = Sigma0 + Sigma1  holds in the simplest case.
    """

    def __init__(self, name, b0, b1, g0, g1, logger=logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.b0name = b0.name
        self.b1name = b1.name
        self.g0 = g0
        self.g1 = g1
        self.record = None
        self._initialize_model()
        self.ts0_recent = -1  # the most recent value of the time_stamp of b0
        self.ts1_recent = -1  # that of b1

        super(MatcherEKF, self).__init__(name, b0, b1)

    def _initialize_model(self):
        self.lnL = 0
        self.err2 = 0

    def __call__(self, inputs):
        """The main routine that is called from brica.
        The input variable  'inputs'  is a python dictionary object that brings all the current states of bundles.
        Ex.

          inputs = {"Bundle0", st0, "Bundle1", st1}

        The output variable 'results' is a python dictionary object which brings all the feedback states to the corresponding bundles.
        Ex.

          results = {"Bundle0", fbst0, "Bundle1", fbst1}

        """
        self.update(inputs)
        return self.results

    def _state_record(self):
        """Storing records of current MatcherEKF
        """
        b0 = self.b0state
        b1 = self.b1state
        fbst0 = self.results[self.b0name]
        fbst1 = self.results[self.b1name]
        # feedback from the current Matcher to the Bundle b1
        mu0 = np.array(b0.data["mu"])
        sigma0 = np.array([np.diag(b0.data["Sigma"])], dtype=np.float32)
        mu1 = np.array(b1.data["mu"])
        sigma1 = np.array([np.diag(b1.data["Sigma"])], dtype=np.float32)
        dmu1 = np.array(fbst1.data["mu"])
        dsigma1 = np.array([np.diag(fbst1.data["Sigma"])], dtype=np.float32)

        if self.record is None:
            self.record = {
                "mu0": mu0,
                "diagSigma0": sigma0,
                "mu1": mu1,
                "diagSigma1": sigma1,
                "dmu1": dmu1,
                "diagDSigma1": dsigma1
            }
            self._first_call_of_state_record = False
        else:
            self.record["mu0"] = np.vstack((self.record["mu0"], mu0))
            self.record["diagSigma0"] = np.concatenate((self.record["diagSigma0"], sigma0), axis=0)
            self.record["mu1"] = np.vstack((self.record["mu1"], mu1))
            self.record["diagSigma1"] = np.concatenate((self.record["diagSigma1"], sigma1), axis=0)
            self.record["dmu1"] = np.vstack((self.record["dmu1"], dmu1))
            self.record["diagDSigma1"] = np.concatenate((self.record["diagDSigma1"], dsigma1), axis=0)

    def forward(self):
        """Main method that evaluates the error and derivatives.
           z = g0(mu0) - g1(mu1)
           C0 = (d g0/ d mu0)   <-- Jacobian
           C1 = (d g1/ d mu1)   <-- Jacobian
           S = C0 Sigma0 C0' + C1 Sigma1 C1'
           err2 = z S^{-1} z'
           lnL = - (1/2) err2 - (1/2) log det( 2*pi*S )
           K0 = Sigma0 C0' S^{-1}
           K1 = Sigma1 C1' S^{-1}
           dmu0 <-- K0 z,  dSigma0 <-- K0 C0 dSigma0
           dmu1 <-- K1 z,  dSigma1 <-- K1 C1 dSigma1

        Note
           When bundle  b0  is a Observer that provides sequencial data, the user may consider the simplest setting with the identity function  g0()  and a fixed noise matrix  Sigma0. In this case, the observer omits the corresponding feedback signal  dmu0  and  dSigma0, and the total covariance  S  becomes that of a standard EKF,
               S = R + C1 Sigma1 C1
           where  R = Sigma0.
           In other words, user may provide a fixed noise matrix  Sigma0 = R  in the observer in order to set the observation noise model.
           Missing observation can be described as temporally setting of large diagonal elements of  Sigma0.
        """
        self.logger.debug("Matcher_EKF forward")
        self.lnL_t = 0
        
        z = self.g0.value(self.mu0) - self.g1.value(self.mu1)
        C0 = self.g0.x(self.mu0)
        C1 = self.g1.x(self.mu1)
        
        S = C0 @ self.Sigma0 @ C0.T + C1 @ self.Sigma1 @ C1.T
        SI = np.linalg.inv(S)
        
        dum_sign, logdet = np.linalg.slogdet(S)
        self.lnL_t -= z @ SI @ z.T / 2.0
        self.err2 += z @ z.T
        self.lnL_t -= logdet / 2.0
        
        K0 = self.Sigma0 @ C0.T @ SI
        K1 = self.Sigma1 @ C1.T @ SI

        self.dmu0 = -K0 @ z
        self.dmu1 =  K1 @ z
        self.dSigma0 = -K0 @ C0 @ self.Sigma0
        self.dSigma1 = -K1 @ C1 @ self.Sigma1
        
        self.lnL += self.lnL_t
        self.logger.debug("lnL_t = {lnLt}, lnL = {lnL}".format(lnLt=self.lnL_t, lnL=self.lnL))

    def backward(self):
        """Updates the observation models
              self.g0 and self.g1
        if they are variables.
        (to be implemented soon)
        """
        self.logger.debug("{} backward".format(self.name))

    def update(self, inputs):
        """method self.update()
         is called from self.__call__()
        """
        self.b0state, self.b1state = inputs[self.b0name], inputs[self.b1name]
        d0, d1 = self.b0state, self.b1state

        self.mu0 = d0["mu"]
        self.Sigma0 = d0["Sigma"]
        self.ts0 = d0["time_id"]
        self.mu1 = d1["mu"]
        self.Sigma1 = d1["Sigma"]
        self.ts1 = d1["time_id"]

        self.logger.debug("mu0={}".format(self.mu0))
        self.logger.debug("Sigma0={}".format(self.Sigma0))
        self.logger.debug("mu1={}".format(self.mu1))
        self.logger.debug("Sigma1={}".format(self.Sigma1))

        if self.ts0 == self.ts0_recent:
            self.logger.debug("b0.state is not updated")
        if self.ts1 == self.ts1_recent:
            self.logger.debug("b1.state is not updated")
        self.ts0_recent = self.ts0
        self.ts1_recent = self.ts1

        self.forward()
        self.backward()

        self.results[self.b0name]["mu"] = self.dmu0.astype(np.float32)
        self.results[self.b0name]["Sigma"] = self.dSigma0.astype(np.float32)
        self.results[self.b1name]["mu"] = self.dmu1.astype(np.float32)
        self.results[self.b1name]["Sigma"] = self.dSigma1.astype(np.float32)



import logging
import numpy as np

from matchernet.log import logging_conf
from matchernet import matchernet

logging_conf.set_logger_config("./log/logging.json")
logger = logging.getLogger(__name__)


class Observer(matchernet.Bundle):
    """Observer works as a Bundle that provides vector data at each time step of the MatcherNet simulation.

    Usage:
    Construct an observer by

    >> o1 = Observer("Observer1", buffer)

    where buffer is an numpy.ndarray of shape (length, dim).
    Then, call it for each step by

    >> result = o1( dummy_input )

    and you get the vector data of the current time-stamp

     results.state["mu"]

    with corresponding observation error covariance matrix

     results.state["Sigma"].

    Note:
    When the vector data in buffer included  NaN  entries, they are regarded as missing entries and the Observer outpus a zero vector  mu  with covariance matrix  cov  of large eigen values. (See the function  missing_handler001()  for a default setting to construct the corresponding output. )
    """

    def __init__(self, name, buff, obs_noise_covariance=None, logger=logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.name = name
        self.buffer = buff
        self.counter = -1
        self.length = buff.shape[0]
        self.dim = buff.shape[1]

        if obs_noise_covariance is None:
            self.obs_noise_covariance = 1000 * np.eye(self.dim, dtype=np.float32)
        else:
            self.obs_noise_covariance = obs_noise_covariance
        self.missing_handler = missing_handler
        # default setting of missing value handler function
        # TODO:
        #self.set_results()
        # set the first value with large obs_noise_covariance
        # for an initial value
        super(Observer, self).__init__(self.name)

    def __call__(self, inputs):
        """ The main routine that is called from brica.
        """
        self.count_up()
        return self.get_results()

    def count_up(self):
        self.counter = (self.counter + 1) % self.length

    def set_buffer(self, buff):
        self.buffer = buff
        self.counter = -1
        self.length = buff.shape[0]
        self.dim = buff.shape[1]

    def get_buffer(self):
        return self.buffer

    def get_state(self):
        b = self.get_buffer()
        z = b[self.counter].copy()
        return z

    def get_results(self):
        q = self.get_state()
        mu, Sigma = self.missing_handler(np.array(q, dtype=np.float32),
                                         self.obs_noise_covariance, self.logger)

        results = {}
        results["mu"] = mu
        results["Sigma"] = Sigma
        #results["time_stamp"] = self.counter
        results["time_id"] = self.counter
        return {
            "state" : results
        }
        # === Note: We may regard  "time_stamp"  as a real time rather than a counter in a future version.


def missing_handler(mu, Sigma, logger=logger):
    """A missing value handler function.
    It receives a vector data  mu  with a default covariance matrix  Sigma, find NaN in the vector  mu, and outputs a modified set of a vector  mu  and a covariance  cov.
    """
    if np.any(np.isnan(mu)):
        logger.debug("Missing!")
        cov = Sigma * 1000
        mu = np.zeros(mu.shape)
    else:
        cov = Sigma
    return mu, cov


class ObserverMultiple(Observer):
    """A bundle that provides sequencial data
    """

    def __init__(self, name, buff, mul):
        self.name = name
        self.mul = mul
        super().__init__(self.name, buff)

    def get_state(self):
        b = self.get_buffer()
        z = b[self.counter].copy()
        for i in range(1, self.mul):
            j = (self.counter + i) % self.length
            z = np.concatenate((z, b[j].copy()))
        return z



# -*- coding: utf-8 -*-
import numpy as np
import brica
from brica import Component

from matchernet.control.ilqg import iLQG
from matchernet.matchernet import Bundle


class MatcherController(object):
    def __init__(self, name, mpcenv_bundle, ekf_bundle, plan_bundle):
        self.name = name
        self.results = {}

        self.mpcenv_bundle = mpcenv_bundle
        self.ekf_bundle = ekf_bundle
        self.plan_bundle = plan_bundle
        
        self.results[self.mpcenv_bundle.name] = {}
        self.results[self.ekf_bundle.name] = {}
        # Note: there is no results entry for plan_bundle

        # TODO: How do we reset this integral value for PID?
        self.integral = 0.0
        self.x_last = None

        self.update_component()

    def update_component(self):
        component = Component(self)

        # MatcherController -> MPCEnvBundle
        component.make_out_port(self.mpcenv_bundle.name)
        self.mpcenv_bundle.component.make_in_port(self.name)  # -> send u to MPCEnv
        brica.connect(component, self.mpcenv_bundle.name,
                      self.mpcenv_bundle.component, self.name)

        # BundleEKFWithController <-> MatcherController
        component.make_in_port(self.ekf_bundle.name)  # receive state from BundleEKF
        component.make_out_port(self.ekf_bundle.name)  # send u to BundleEKF
        self.ekf_bundle.component.make_in_port(self.name)

        brica.connect(self.ekf_bundle.component, "state", 
                      component, self.ekf_bundle.name)
        brica.connect(component, self.ekf_bundle.name,
                      self.ekf_bundle.component, self.name)

        # BundlePlan -> MatcherController
        component.make_in_port(self.plan_bundle.name)  # receive state from BundlePlan
        brica.connect(self.plan_bundle.component, "state", 
                      component, self.plan_bundle.name)
        
        self.component = component

    def __call__(self, inputs):
        self.update(inputs)

        return self.results

    def update(self, inputs):
        ekf_state = inputs[self.ekf_bundle.name]
        plan_state = inputs[self.plan_bundle.name]
        
        x_plan = plan_state["x"]
        u_plan = plan_state["u"]
        K_plan = plan_state["K"]
        K_i_plan = plan_state["K_i"]
        K_d_plan = plan_state["K_d"]
        eta = plan_state["eta"]
        dt = plan_state["dt"]
        plan_time_stamp = plan_state["time_stamp"]
        plan_time_id = plan_state["time_id"]

        mu = ekf_state["mu"]
        ekf_time_stamp = ekf_state["time_stamp"]
        ekf_time_id = ekf_state["time_id"]

        # PID control
        u_p = u_plan + K_plan @ (mu - x_plan)
        self.integral = self.integral * (1 - eta * dt) + eta * dt * (mu - x_plan)
        u_i = -K_i_plan @ self.integral
        if self.x_last is None:
            self.x_last = np.copy(mu)
        u_d = -K_d_plan @ (mu - self.x_last) / dt
        u = u_p + u_i + u_d
        
        self.x_last = np.copy(mu)
        
        self.results[self.mpcenv_bundle.name]["u"] = u
        self.results[self.ekf_bundle.name]["u"] = u


class Plan(object):
    def __init__(self, x_list, u_list, K_list, time_stamp, time_id):
        self.x_list = x_list
        self.u_list = u_list
        self.K_list = K_list
        self.time_stamp = time_stamp
        self.time_id = time_id


class MatcherILQR(object):
    def __init__(self, name, ekf_bundle, plan_bundle,
                 dynamics, cost, dt, T, iter_max):
        self.name = name
        self.results = {}

        self.ekf_bundle = ekf_bundle
        self.plan_bundle = plan_bundle
        self.bundles = (ekf_bundle, plan_bundle)

        self.results[self.plan_bundle.name] = {}

        self.ilqg = iLQG(dynamics=dynamics, cost=cost, dt=dt)
        
        self.T = T
        self.iter_max = iter_max

        self.last_u_list = None
        self.last_time_id = 0
        
        self.update_component()

    def update_component(self):
        component = Component(self)

        # BundleEKFWithController -> MatcherILQR
        component.make_in_port(self.ekf_bundle.name)  # receive state from BundleEKF
        brica.connect(self.ekf_bundle.component, "state", component,
                      self.ekf_bundle.name)

        # MatcherController -> BundlePlan
        component.make_out_port(self.plan_bundle.name)  # send Plan to BundlePlan
        self.plan_bundle.component.make_in_port(self.name)
        brica.connect(component, self.plan_bundle.name,
                      self.plan_bundle.component, self.name)
        
        self.component = component

    def __call__(self, inputs):
        self.update(inputs)

        return self.results

    def update(self, inputs):
        ekf_state = inputs[self.ekf_bundle.name]
        
        mu = ekf_state["mu"]
        time_stamp = ekf_state["time_stamp"]
        time_id = ekf_state["time_id"]

        # Initial control sequence for MPC
        u_dim = self.ilqg.dynamics.u_dim
        u0 = np.zeros((self.T, u_dim), dtype=np.float32)
        if self.last_u_list is not None:
            # Set initial control signals by copying last control signals.
            advance_time_id = time_id - self.last_time_id
            u0[:self.T-advance_time_id,:] = self.last_u_list[advance_time_id:,:]

        x_list, u_list, K_list = self.ilqg.optimize(mu,
                                                    u0,
                                                    self.T,
                                                    self.iter_max)
        plan = Plan(x_list, u_list, K_list, time_stamp, time_id)
        self.results[self.plan_bundle.name]["plan"] = plan
        
        self.last_u_list = u_list
        self.last_time_id = time_id


class BundlePlan(Bundle):
    """
    Plan sending class for iLQR control.
    """
    def __init__(self, name, x_dim, u_dim, dt, control_T, plan_src_name):
        super(BundlePlan, self).__init__(name)
        self.x_dim = x_dim
        self.u_dim = u_dim
        self.dt = dt
        self.control_T = control_T
        self.plan_src_name = plan_src_name
        
        self.latest_plan = None
        self.index_in_plan = 0
        
        self.update_component()

    def __call__(self, inputs):
        if self.plan_src_name in inputs.keys() and inputs[self.plan_src_name] is not None:
            plan = inputs[self.plan_src_name]["plan"]
            if (self.latest_plan is None) or (self.latest_plan.time_id != plan.time_id):
                self.latest_plan = plan
                self.index_in_plan = self.control_T - 1
                # Note: To make the timestamp equal to that from EKFBundle, we need to start
                # index_in_plan as 'control_T + 1', but we are now using 'control_T - 1'
        
        if self.latest_plan is not None:
            x = self.latest_plan.x_list[self.index_in_plan]
            u = self.latest_plan.u_list[self.index_in_plan]
            K = self.latest_plan.K_list[self.index_in_plan]
            time_stamp = self.latest_plan.time_stamp + self.dt * self.index_in_plan
            time_id = self.latest_plan.time_id + self.index_in_plan
            self.index_in_plan += 1
        else:
            x = np.zeros((self.x_dim,), dtype=np.float32)
            u = np.zeros((self.u_dim,), dtype=np.float32)
            K = np.zeros((self.u_dim, self.x_dim), dtype=np.float32)
            time_stamp = 0
            time_id = 0

        # Compatibility for PID controller
        K_i = np.zeros((self.u_dim, self.x_dim), dtype=np.float32)
        K_d = np.zeros((self.u_dim, self.x_dim), dtype=np.float32)
        eta = 0.0
        
        results = {
            "x": x,
            "u": u,
            "K": K,
            "K_i": K_i,
            "K_d": K_d,
            "eta": eta,
            "dt": self.dt,
            "time_stamp": time_stamp,
            "time_id": time_id
        }
        return {"state": results}


class BundleFixedPlan(Bundle):
    """
    Plan sending class for PID control.
    """
    def __init__(self, name, x_dim, u_dim, dt, x_target, K, K_i, K_d, eta):
        super(BundleFixedPlan, self).__init__(name)
        self.x_dim = x_dim
        self.u_dim = u_dim
        self.x_target = x_target
        self.K = K
        self.K_i = K_i
        self.K_d = K_d
        self.eta = eta
        self.dt = dt

        self.time_stamp = 0.0
        self.time_id = 0
        
        self.update_component()

    def __call__(self, inputs):
        # Compatibility for iLQR controller
        u = np.zeros((self.u_dim,), dtype=np.float32)
        
        results = {
            "x": self.x_target,
            "u": u,
            "K": -self.K, # Sign is inverted for the compatibility with iLQR
            "K_i": self.K_i,
            "K_d": self.K_d,
            "eta": self.eta,
            "dt": self.dt,
            "time_stamp": self.time_stamp,
            "time_id": self.time_id
        }

        self.time_stamp += self.dt
        self.time_id += 1
        
        return {"state": results}


class MPCEnvBundle(Bundle):
    """
    MPCEnv bundle class that communicates with matcher.
    """

    def __init__(self, name, env, R, control_src_name, debug_recorder=None):
        """
        Arguments:
          env
            MPCEnv instance
          R
            Observation noise
          control_src_name
            Name of controller src node (Controller Matcher)
        """
        super(MPCEnvBundle, self).__init__(name)
        self.env = env
        self.R = R
        self.time_stamp = 0.0
        self.time_id = 0
        self.control_src_name = control_src_name
        self.debug_recorder = debug_recorder
        
        self.update_component()

    def __call__(self, inputs):
        u_dim = self.env.u_dim

        # Receive action from Controller Matcher
        if self.control_src_name in inputs.keys() and inputs[self.control_src_name] is not None:
            u = inputs[self.control_src_name]["u"]
        else:
            u = np.zeros((u_dim,), dtype=np.float32)

        # Step environment with received action
        # Ignoring env rewards
        x, _ = self.env.step(u)
        
        results = {}
        results["mu"] = x
        results["Sigma"] = self.R
        results["time_stamp"] = self.time_stamp
        results["time_id"] = self.time_id

        if self.debug_recorder is not None:
            self.debug_recorder.record(x, u)

        self.time_stamp += self.env.dt
        self.time_id += 1

        # Send state to matcher
        return {"state": results}



# -*- coding: utf-8 -*-
import numpy as np
import os
import shutil
import cv2
import subprocess


class MovieWriter(object):
    """ Movie output utility class. """
    def __init__(self, file_name, frame_size, fps):
        """
        frame_size is (w, h)
        """
        self._frame_size = frame_size
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        self.vout = cv2.VideoWriter()
        success = self.vout.open(file_name, fourcc, fps, frame_size, True)
        if not success:
            print("Create movie failed: {0}".format(file_name))

    def add_frame(self, frame):
        """
        frame shape is (h, w, 3), dtype is np.uint8
        """
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        self.vout.write(frame)

    def close(self):
        self.vout.release()
        self.vout = None


class AnimGIFWriter(object):
    """ Anim GIF output utility class. """
    def __init__(self, file_name, fps, tmp_dir_name="tmp_gif"):
        """
        frame_size is (w, h)
        """
        self.file_name = file_name
        self.fps = fps
        self.tmp_dir_name = tmp_dir_name
        self.image_index = 0

        if os.path.exists(self.tmp_dir_name):
            shutil.rmtree(self.tmp_dir_name)
        os.mkdir(self.tmp_dir_name)

    def add_frame(self, frame):
        """
        frame shape is (h, w, 3), dtype is np.uint8
        """
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        file_path = os.path.join(self.tmp_dir_name, "frame{0:03d}.png".format(self.image_index))
        cv2.imwrite(file_path, frame)
        self.image_index += 1
        
    def close(self):
        delay = int(round(100 / self.fps))
        command = "convert -delay {} {}/*.png {}".format(delay,
                                                         self.tmp_dir_name,
                                                         self.file_name)
        subprocess.call(command.split())
        shutil.rmtree(self.tmp_dir_name)



"""
matchernet.py
=====

This module contains a null demonstration of the BundleNet

"""
import logging
from operator import add
from functools import reduce

from matchernet.log import logging_conf
from matchernet.matchernet import Bundle, Matcher
from matchernet import state

logging_conf.set_logger_config("./log/logging.json")
logger = logging.getLogger(__name__)


class BundleNull(Bundle):
    def __init__(self, name, mu, delay=0, logger=logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.name = name
        # self.delay = delay
        self.delay = -1
        self.state = state.StatePlain(mu)
        super(BundleNull, self).__init__(self.name)

    def accept_feedback(self, fb_state):
        self.state.data["mu"] += fb_state["mu"]

    def __call__(self, inputs):
        for key in inputs:
            if inputs[key] is not None:
                self.accept_feedback(inputs[key])

        self.update(inputs)
        results = {}
        results["mu"] = self.state.data["mu"]
        return {
            "state": results
        }

    def update(self, inputs):
        self.logger.debug("Updating {}".format(self.name))
        # if(self.delay):
        # ref = time.clock_gettime(time.CLOCK_MONOTONIC)
        # elapsed = 0
        # while elapsed < self.delay:
        # elapsed = time.clock_gettime(time.CLOCK_MONOTONIC) - ref


class MatcherNull2(Matcher):
    """MatcherNull is a Matcher that connects two BundleNull s
    """

    def __init__(self, name, b0, b1):
        self.name = name
        super(MatcherNull2, self).__init__(self.name, b0, b1)

    def update(self, inputs):
        mu = {}
        for key in inputs:
            mu[key] = inputs[key]["mu"]
        mean = reduce(add, mu.values()) / len(inputs)
        for key in inputs:
            if inputs[key] is not None:
                self.results[key]["mu"] = (mean - mu[key]) * 0.1



"""
fn.py
=====

This module contains function handler classes that the BundleNet
architecture needs. It is overridden when you use chainer/TensorFlow to implement the arbitrary parametric functions.

"""

import autograd.numpy as np
from autograd import jacobian


class Fn(object):
    """An abstract class to implement numerical function
    that BundleNet uses.
    """

    def __init__(self, A):
        self.A = A
        self.x = jacobian(self.value, 0)

    def get_params(self):
        return self.A

    def value(self, x):
        """Numerically calculates f(x)
          x should be a numpy array of shape (dim_in, 1)
          outputs a numpy array of shape (dim_out, 1)
        """
        return 0


class LinearFn(Fn):
    """Linear function y = np.dot(A, x) and its derivatives.
    """

    def __init__(self, A):
        super(LinearFn, self).__init__(A)
        self.A = A
        self.x = jacobian(self.value, 0)

    def value(self, x):
        return self.A @ x


class LinearFnXU(object):
    """Linear function y = np.dot(A, x) + np.dot(B, u) and its derivatives.

    .. note:: The shapes of matrix A and matrix B must match

    """

    def __init__(self, A, B):
        self.A = A
        self.B = B
        self.x = jacobian(self.value, 0)
        self.u = jacobian(self.value, 1)

    def value(self, x, u):
        return self.A @ x + self.B @ u



"""
matchernet.py
=====

This module contains basic abstract classes that the BundleNet
architecture needs; 'Bundle' and 'Matcher'. It also includes
their tests , function 'test_abstract_bm'.
It imports BriCA2 (impremented with C++)
https://github.com/BriCA/BriCA2

"""
import logging
import brica
from brica import Component

from matchernet.log import logging_conf

logging_conf.set_logger_config("./log/logging.json")
logger = logging.getLogger(__name__)


class Bundle(object):
    """
    'Bundle' is an abstract class that defines basic propaties of Bundles in a matchernet.

    Each bundle has
      a state
      a state transision dynamics
      arbitrary number of connections to Matchers
    """

    def __init__(self, name, logger=logger):
        """ Create a new 'Bundle' instance.
        """
        self.logger = logger.getChild(self.__class__.__name__)
        self.name = name

        self.update_component()

    def update_component(self):
        self.component = Component(self)
        self.component.make_out_port("state")

    def __call__(self, inputs):
        """ The main routine that is called from brica.
        """
        for key in inputs:  # key is one of the matcher names
            if inputs[key] is not None:
                self.accept_feedback(inputs[key])

        self.update(inputs)
        state = {}
        return {"state": state}

    def accept_feedback(self, fb_state):
        """
        Accepting the feedback state 'fb_state' from one of the matchers
        which is linking to the current bundle.
        """
        self.logger.debug("{} is accepting feedback".format(self.name))

    def update(self, inputs):
        """ Update the state of the current Bundle.
        This method should be override for any subclasses.
        """
        self.logger.debug("{} is updating".format(self.name))


class Matcher(object):
    """
    'Matcher' is an abstract class that defines basic propaties of Matchers in a matchernet.
    Matcher is a component of matchernet.
    It connects Bundles, calculates energy, and returns a feedback state for each Bundle.
    Here, energy stands for a measure of contradiction among all the linking Bundles. The feedback stands for a signal that is used at the corresponding Bundle; the Bundle updates its state in order to decrease the energy.

    Args:
        name (str): matcher name.
        *bundles (:class:`~chainer.Variable`):
    """

    def __init__(self, name, *bundles, logger=logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.name = name
        self.results = {}
        self.bundles = bundles
        for bundle in self.bundles:
            self.results[bundle.name] = {}
        self.update_component()

    def update_component(self):
        self.component = Component(self)
        for bundle in self.bundles:
            self.component.make_in_port(bundle.name)
            self.component.make_out_port(bundle.name)
            bundle.component.make_in_port(self.name)

            brica.connect(bundle.component, "state", self.component, bundle.name)
            brica.connect(self.component, bundle.name, bundle.component, self.name)

    def __call__(self, inputs):
        """
        The main routine that is called from brica.
        The input variable  'inputs'  is a python dictionary object that brings all the current states of bundles.
        Ex.

          inputs = {"Bundle0", st0, "Bundle1", st1}

        The output variable 'results' is a python dictionary object which brings all the feedback states to the corresponding bundles.
        Ex.

          results = {"Bundle0", fbst0, "Bundle1", fbst1}

        """
        for key in inputs:
            if inputs[key] is not None:
                self.accept_bundle_state(inputs[key])
        self.update(inputs)
        return self.results

    def accept_bundle_state(self, st):
        self.logger.debug("Matcher {} is accepting_bundle_state".format(self.name))

    def update(self, inputs):
        self.logger.debug("self.results={x}".format(x=self.results))



import numpy as np

from matchernet import utils


class StateSpaceModel2Dim(object):
    def __init__(self, n_dim, A, g, sigma_w, sigma_z, x, y):
        if n_dim != 2:
            raise NotImplementedError
        self.n_dim = n_dim
        self.A = A
        self.g = g
        self.sigma_w = sigma_w
        self.sigma_z = sigma_z
        self.x = x
        self.y = y

    def simulation(self, n_steps, dt):
        F = utils.calc_matrix_F(self.A, dt)
        for i in range(n_steps):
            (x, y) = self._step(F)
            if i == 0:
                x_series = x
                y_series = y
            else:
                x_series = np.concatenate((x_series, x), axis=0)
                y_series = np.concatenate((y_series, y), axis=0)

        return x_series, y_series

    def _step(self, F):
        self.w = self.sigma_w * np.random.standard_normal(size=(1, self.n_dim))
        self.x = np.dot(self.x, F) + self.w
        self.z = self.sigma_z * np.random.standard_normal(size=(1, self.n_dim))
        self.y = self.x + self.z

        return self.x, self.y



# -*- coding: utf-8 -*-
import numpy as np

from matchernet import utils


class Regularization:
    def __init__(self):
        self.lambd         = 0.1
        self.d_lambda      = 1.0
        self.lambda_factor = 1.6
        self.lambda_max    = 1e10
        self.lambda_min    = 1e-6
    
    def on_diverge(self):
        self.d_lambda = max(self.d_lambda * self.lambda_factor, self.lambda_factor)
        self.lambd = max(self.lambd * self.d_lambda, self.lambda_min)
        self.lambd = min(self.lambd, self.lambda_max)

    def get_current_lamba(self):
        return self.lambd

    

class iLQG(object):
    def __init__(self, dynamics, cost, dt):
        self.dynamics = dynamics
        self.stored_cost = cost
        self.dt = dt

    def optimize(self,
                 x_init,
                 u_init,
                 T,
                 start_time_step=0,
                 iter_max=20,
                 stop_criteria=1e-3,
                 u_min=None,
                 u_max=None):

        assert x_init.shape == (self.dynamics.x_dim,)
        assert u_init.shape == (T, self.dynamics.u_dim)
        
        self.start_time_step = start_time_step
        
        # Clone cost
        cost = self.stored_cost.clone()
        
        x_list = self.init_trajectory(x_init, u_init, T, cost)
        
        u_list = np.copy(u_init)

        self.regularization = Regularization()
        
        for i in range(iter_max):
            #print("iter={}".format(i))
            
            for j in range(10):
                lambd = self.regularization.get_current_lamba()
                diverged, k_list, K_list = self.backward(x_list, u_list, T, cost, lambd)
                if not diverged:
                    break
                else:
                    self.regularization.on_diverge()
                    print("update lambda={}".format(
                        self.regularization.get_current_lamba()))

            # Clone cost
            cost = self.stored_cost.clone()
            
            x_list, u_list, diff = self.forward(x_list, u_list, x_init,
                                                k_list, K_list, T, cost,
                                                u_min, u_max)

            if(diff < stop_criteria):
                print("it={}, diff={}".format(i, diff))
                break
            
        return np.array(x_list, dtype=np.float32), \
            np.array(u_list, dtype=np.float32), \
            np.array(K_list, dtype=np.float32)
    
    def init_trajectory(self, x_init, u_init, T, cost):
        x_list = [x_init]
        
        for i in range(T):
            cost.apply_state(x_list[i], (self.start_time_step + i) * self.dt)
            xdot = self.dynamics.value(x_list[i], u_init[i])
            next_x = x_list[i] + xdot * self.dt
            x_list.append(next_x)
            
        return x_list

    def forward(self, x_list, u_list, x_init, k_list, K_list, T, cost, u_min, u_max):
        next_x_list = [x_init]
        next_u_list = []
        
        diff = 0.0
        alpha = 1.0

        for t in range(T):
            cost.apply_state(next_x_list[t], (self.start_time_step + t) * self.dt)

            # TODO:
            next_u = u_list[t] + alpha * k_list[t] + \
                     K_list[t] @ (next_x_list[t] - x_list[t])

            if u_min is not None:
                next_u = np.max([next_u, np.ones_like(next_u) * u_min], axis=0)
            if u_max is not None:
                next_u = np.min([next_u, np.ones_like(next_u) * u_max], axis=0)
            
            next_u_list.append(next_u)
            xdot = self.dynamics.value(next_x_list[t], next_u_list[t])
            next_x = next_x_list[t] + xdot * self.dt
            next_x_list.append(next_x)
            
            diff += np.sum(np.abs(u_list[t] - next_u_list[t]))
        
        return next_x_list, next_u_list, diff

    def backward(self, x_list, u_list, T, cost, lambd):
        k_list = []
        K_list = []

        # Derivatives of the terminal cost
        lx  = cost.x( x_list[T], None, self.start_time_step * self.dt)
        lxx = cost.xx(x_list[T], None, self.start_time_step * self.dt)
        
        Vx  = lx  # (x_dim,)
        Vxx = lxx # (x_dim,x_dim)
        
        assert Vx.shape  == (self.dynamics.x_dim,)
        assert Vxx.shape == (self.dynamics.x_dim, self.dynamics.x_dim)

        diverged = False
        
        for t in range(T-1, -1, -1):
            x = x_list[t]
            u = u_list[t]
            
            assert x.shape == (self.dynamics.x_dim,)
            assert u.shape == (self.dynamics.u_dim,)
            
            # Derivatives of the dynamics
            fx = self.dynamics.x(x, u)
            fu = self.dynamics.u(x, u)
            
            Fx = utils.calc_matrix_F(fx, self.dt)
            Fu = fu * self.dt
            
            # Derivatives of the running cost
            lx  = cost.x( x, u, (self.start_time_step+t) * self.dt)
            lxx = cost.xx(x, u, (self.start_time_step+t) * self.dt)
            lu  = cost.u( x, u, (self.start_time_step+t) * self.dt)
            luu = cost.uu(x, u, (self.start_time_step+t) * self.dt)
            lux = cost.ux(x, u, (self.start_time_step+t) * self.dt)
            
            # Derivatives of the Q function
            Qx  = lx  + Fx.T @ Vx
            Qu  = lu  + Fu.T @ Vx
            Qxx = lxx + Fx.T @ Vxx @ Fx
            Quu = luu + Fu.T @ Vxx @ Fu
            Qux = lux + Fu.T @ Vxx @ Fx
            
            # Regularlize
            Quu = Quu + np.eye(self.dynamics.u_dim) + lambd
            
            assert Qx.shape  == (self.dynamics.x_dim,)
            assert Qu.shape  == (self.dynamics.u_dim,)
            assert Qxx.shape == (self.dynamics.x_dim, self.dynamics.x_dim)
            assert Quu.shape == (self.dynamics.u_dim, self.dynamics.u_dim)
            assert Qux.shape == (self.dynamics.u_dim, self.dynamics.x_dim)

            diverged, Quu_inv = self.calc_inverse(Quu)
            if diverged:
                print("Diverged")
                break
            
            k = -Quu_inv @ Qu
            K = -Quu_inv @ Qux
            
            assert k.shape == (self.dynamics.u_dim,)
            assert K.shape == (self.dynamics.u_dim, self.dynamics.x_dim)

            Vx  = Qx  + K.T @ Quu @ k + K.T @ Qu  + Qux.T @ k
            Vxx = Qxx + K.T @ Quu @ K + K.T @ Qux + Qux.T @ K
            Vxx = 0.5 * (Vxx + Vxx.T) # Fix symmetry
            
            assert Vx.shape  == (self.dynamics.x_dim,)
            assert Vxx.shape == (self.dynamics.x_dim, self.dynamics.x_dim)
            
            k_list.append(k)
            K_list.append(K)
            
        k_list.reverse()
        K_list.reverse()

        return diverged, k_list, K_list

    def calc_inverse(self, m):
        try:
            R = np.linalg.cholesky(m)
        except np.linalg.LinAlgError:
            return True, None
        R_inv = np.linalg.inv(R)
        m_inv = R_inv.T @ R_inv
        return False, m_inv



# -*- coding: utf-8 -*-

class Dynamics(object):
    """
    Base interfaces for MPC dynamics definition.
    """
    def __init__(self):
        pass

    def value(self, x, u):
        raise NotImplementedError()

    def x(self, x, u):
        raise NotImplementedError()

    def u(self, x, u):
        raise NotImplementedError()

    @property
    def x_dim(self):
        raise NotImplementedError()

    @property
    def u_dim(self):
        raise NotImplementedError()


class Renderer(object):
    """
    Base interfaces for MPC agent renderer.
    """
    def __init__(self):
        pass

    def render(self, x, u=None, override_image=None):
        raise NotImplementedError()


class RewardSystem(object):
    """
    Base interfaces system class for MPC environment.
    """
    def __init__(self):
        pass

    def reset(self):
        raise NotImplementedError()

    def render(self, image):
        raise NotImplementedError()

    def evaluate(self, x, dt):
        raise NotImplementedError()


class Cost(object):
    def __init__(self):
        pass

    def clone(self):
        return self

    def apply_state(self, x, t):
        pass

    def value(self, x, u, t):
        raise NotImplementedError()



from matchernet.ekf import *
from matchernet.fn import *
from matchernet.matchernet import *
from matchernet.matchernet_null import *
from matchernet.observer import *
from matchernet.state import *
from matchernet.state_space_model_2d import *
from matchernet.utils import *
from matchernet.control.mpcenv import *
from matchernet.control.control import *
from matchernet.control.mpc import *
from matchernet.control.multi import *
from matchernet.control.ilqg import *
from matchernet import *
from matchernet.misc.recording import *

__copyright__ = 'Copyright (C) 2019 Shigeyuki Oba'
__version__ = '0.0.12'
__license__ = 'Apache License 2.0'
__author__ = 'Shigeyuki Oba'
__author_email__ = 'oba@i.kyoto-u.ac.jp'
__url__ = 'https://github.com/shigeyukioba/matchernet'
__all__ = [
    "ekf",
    "fn",
    "matchernet",
    "matchernet_null",
    "observer",
    "state",
    "state_space_model_2d",
    "utils"
    "control"
]



# -*- coding: utf-8 -*-
import numpy as np


class MPCEnv(object):
    def __init__(self,
                 dynamics,
                 renderer,
                 reward_system,
                 dt,
                 use_visual_state=False):
        """
        Arguments:
          dynamics:
             Agent dynamics
          renderer:
             Renderer or list of Renderer (Agent renderer)
          reward_system:
             RewardSystem or None
          dt:
             Timestep
          use_visual_state:
             Whether to use visual state output or not (bool)
        """
        self.dynamics = dynamics
        self.renderer = renderer
        self.reward_system = reward_system
        self.dt = dt
        self.use_visual_state = use_visual_state
        self.reset()
        
    def reset(self, x_init=None):
        """
        Reset the environment:

        Arguments:
          x_init:
             Initial state (can be None)

        Returns:
          Current state after resetting
        """
        if x_init is None:
            self.x = np.zeros((self.dynamics.x_dim,), dtype=np.float32)
        else:
            self.x = np.copy(x_init)
        if self.reward_system is not None:
            self.reward_system.reset()
        return self.get_state(action=np.zeros((self.dynamics.u_dim,)))
    
    def get_state(self, action):
        if self.use_visual_state:
            return self.get_visual_state(action)
        else:
            return self.x
        
    def get_visual_state(self, action):
        if isinstance(self.renderer, list):
            # For multi angle view rendering
            renderer_size = len(self.renderer)
            images = []
            for i in range(renderer_size):
                # Render control object
                image = self.renderer[i].render(self.x, action)
                
                # Render rewards
                if self.reward_system is not None:
                    self.reward_system.render(image)
                images.append(image)
            return np.stack(images)
        else:
            # For single angle view rendering
            # Render control object
            image = self.renderer.render(self.x, action)
            
            # Render rewards
            if self.reward_system is not None:
                self.reward_system.render(image)
            
            return image

    def step(self, action):
        """
        Step forward the environment.

        Arguments:
          action
            Control signal
        
        Returns:
          (state, reward)
        """
        xdot = self.dynamics.value(self.x, action)
        self.x = self.x + xdot * self.dt
        
        # Calculate reward
        if self.reward_system is not None:
            reward = self.reward_system.evaluate(self.x, self.dt)
        else:
            reward = 0.0
        return self.get_state(action), reward

    @property
    def u_dim(self):
        return self.dynamics.u_dim

    @property
    def x_dim(self):
        return self.dynamics.x_dim



import numpy as np

from matchernet import utils


class State(object):
    """Class State is a state handler that maintains, serializes, and deserializes the state of Bundles or Matchers.
    The methods serialize() and deserialize() are required for BriCA1 components to exchange their states as numpy.array objects.

    A Bundle/Matcher has its state as a dictionary.
    For exmaple,
    B0.state = state.State()
    B0.state.data = {"mu":np.array([1,2]),
               "Sigma":np.array([[1,0],[0,1]])}
    The disctionary is serialized with a method  serialize()
    q = B0.serialize()
    into a numpy.array object  q  .
    The serialized array is exchanged through  BriCA1 IN/OUT ports
    and deserialized with a method  deserialize() as.
    B0.deserialize(q)
    """

    def __init__(self, mu):
        self.data = {"mu": mu}


class StatePlain(State):
    """StatePlain is a State that handles plain numpy.array.
    """

    def __init__(self, mu):
        """Initializer takes a dimensionarity of the vector.
        """
        super(StatePlain, self).__init__(mu)


class StateMuSigmaDiag(State):
    """StateMuSigmaDiag is a state handler that handles
    state variable as the following dictionary style.
    B.state.data = {"id":1,
            "mu":numpy.array([1,2,3]),
            "sigma":numpy.array([2.0,2.0,2.0])}
    Note that StateMuSigma and StateMuSigmaDiag have
        n x n matrix "Sigma" and n vector "sigma",
        respectively.
    """

    def __init__(self, mu, sigma):
        super(StateMuSigmaDiag, self).__init__(mu)
        self.data["sigma"] = sigma


class StateMuSigma(State):
    """StateMuSigma is a state handler that handles
    state variable as the following dictionary style.
    B.state.data = {"id":1,
            "mu":numpy.array([1,2,3]),
            "Sigma":numpy.array([[1,0,0],[0,1,0],[0,0,1]])}
    Note that StateMuSigma and StateMuSigmaDiag have
        n x n matrix "Sigma" and n vector "sigma",
        respectively.
    """

    def __init__(self, mu, Sigma):
        super(StateMuSigma, self).__init__(mu)
        self.data["Sigma"] = Sigma



import sys
import numpy as np


def print_flush(s):
    sys.stdout.write("\r" + s)
    sys.stdout.flush()


def zeros(size):
    return np.zeros(size).astype(np.float32)


def regularize_cov_matrix(a, mineig=1e-5):
    """
    regularize the covariance matrix  a  so that its minimum eigen value is larger than  mineig.
    """
    l, p = np.linalg.eigh(a)
    # Note: using eigh() rather than eig()
    n = l.size
    for i in range(n):
        if l[i] < mineig:
            l[i] = mineig
    ar = np.dot(np.dot(p, np.diag(l)), p.T)
    return ar


def calc_matrix_F(A, dt):
    """Compute the F matrix.

    Ft' = exp(dt * At') ≈ I + dt * At' + dt^2 * At'^2 + ...
    Approximate up to the third-order term and calculate F.
    F = I + A * dt + (A^2 * dt^2)/2.0 + (A^3 * dt^3)/6.0

    Args:
        A (np.array): Jacobian of the state space model dynamics function.
        dt (int or float): Differentiating time width.

    Returns:
        np.array: A variable holding a np.array of the F matrix.
    """
    n = A.shape[1]
    A2 = np.dot(A, A)
    A3 = np.dot(A2, A)
    F = np.identity(n, dtype=np.float32) + dt * A + dt ** 2 * A2 / 2.0 + dt ** 3 * A3 / 6.0
    return F



def set_logger_config(file_path):
    from json import load
    from logging import config, Filter

    class VisualizeFilter(Filter):

        def __init__(self, words=None):
            if not isinstance(words, list):
                words = None
            self.words = words

        def word_filter(self, record):
            if self.words is not None:
                for word in self.words:
                    if word in record.msg:
                        record.msg = "Filterd: " + word
                        break
            return True

    with open("logging.json", "r", encoding="utf-8") as f:
        log_conf_dic = load(f)

    log_conf_dic["filters"]["visualizeFilter"]["()"] = VisualizeFilter

    config.dictConfig(log_conf_dic)



from setuptools import setup, find_packages
from codecs import open
from os import path
import re

package_name = "matchernet"

root_dir = path.abspath(path.dirname(__file__))


def _requirements():
    return open(path.join(root_dir, 'requirements.txt')).read().splitlines()


with open(path.join(root_dir, package_name, '__init__.py')) as f:
    init_text = f.read()
    version = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    license = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

assert version
assert license
assert author
assert author_email
assert url

with open('readme.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=package_name,
    packages=find_packages(exclude=('tests', )),
    version=version,
    license=license,
    install_requires=_requirements(),
    include_package_data=True,
    author=author,
    author_email=author_email,
    url=url,
    description='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='kalman filter, neural net',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
import autograd.numpy as np
import copy


class MultiAgentDynamics(object):
    """
    Dynamics wrapper for Multi Agent environment.
    """
    def __init__(self, dynamics, agent_size):
        self.dynamics = dynamics
        self.agent_size = agent_size

    def value(self, x, u):
        xdot_new = np.zeros_like(x)
        x_dim = self.dynamics.x_dim
        u_dim = self.dynamics.u_dim
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            u_agent = u[u_dim*i:u_dim*(i+1)]
            xdot_agent = self.dynamics.value(x_agent, u_agent)
            xdot_new[x_dim*i:x_dim*(i+1)] = xdot_agent
        return xdot_new

    def x(self, x, u):
        dx_new = np.zeros((self.x_dim, self.x_dim), dtype=np.float32)
        x_dim = self.dynamics.x_dim
        u_dim = self.dynamics.u_dim
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            u_agent = u[u_dim*i:u_dim*(i+1)]
            dx_agent = self.dynamics.x(x_agent, u_agent) # x_dim, x_dim
            dx_new[x_dim*i:x_dim*(i+1), x_dim*i:x_dim*(i+1)] = dx_agent
        return dx_new
        
    def u(self, x, u):
        du_new = np.zeros((self.x_dim, self.u_dim), dtype=np.float32)
        x_dim = self.dynamics.x_dim
        u_dim = self.dynamics.u_dim
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            u_agent = u[u_dim*i:u_dim*(i+1)]
            du_agent = self.dynamics.u(x_agent, u_agent) # (x_dim, u_dim)
            du_new[x_dim*i:x_dim*(i+1), u_dim*i:u_dim*(i+1)] = du_agent
        return du_new

    @property
    def x_dim(self):
        return self.dynamics.x_dim * self.agent_size

    @property
    def u_dim(self):
        return self.dynamics.u_dim * self.agent_size


class MultiAgentCost(object):
    def __init__(self, cost, agent_size):
        # TODO: costが状態を持つのをやめた方がいいだろう
        self.costs = [cost.clone() for i in range(agent_size)]
        self.agent_size = agent_size
        
    def clone(self):
        # TODO: costが状態を持つのをやめた方がいいだろう
        new_costs = [cost.clone() for cost in self.costs]
        new_multi_cost = copy.copy(self)
        new_multi_cost.costs = new_costs
        return new_multi_cost

    def apply_state(self, x, t):
        # TODO:
        x_dim = x.shape[0] // self.agent_size
        
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            self.costs[i].apply_state(x_agent, t)
    
    def value(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size
        
        value_new = 0.0
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            # TODO:
            value_agent = self.costs[i].value(x_agent, u_agent, t)
            value_new += value_agent
        return value_new

    def x(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size

        dx = np.zeros((x.shape[0],))
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            dx_agent = self.costs[i].x(x_agent, u_agent, t)
            dx[x_dim*i:x_dim*(i+1)] = dx_agent
        return dx

    def u(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size

        du = np.zeros((u.shape[0],))
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            du_agent = self.costs[i].u(x_agent, u_agent, t)
            du[u_dim*i:u_dim*(i+1)] = du_agent
        return du

    def xx(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size

        dxx = np.zeros((x.shape[0],x.shape[0]))
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            dxx_agent = self.costs[i].xx(x_agent, u_agent, t)
            dxx[x_dim*i:x_dim*(i+1),x_dim*i:x_dim*(i+1)] = dxx_agent
        return dxx

    def uu(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size

        duu = np.zeros((u.shape[0],u.shape[0]))
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            duu_agent = self.costs[i].uu(x_agent, u_agent, t)
            duu[u_dim*i:u_dim*(i+1),u_dim*i:u_dim*(i+1)] = duu_agent
        return duu

    def ux(self, x, u, t):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size

        dux = np.zeros((u.shape[0],x.shape[0]))
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None
            dux_agent = self.costs[i].ux(x_agent, u_agent, t)
            dux[u_dim*i:u_dim*(i+1),x_dim*i:x_dim*(i+1)] = dux_agent
        return dux



class MultiAgentRenderer(object):
    """
    Renderer wrapper for Multi Agent environment.
    """
    def __init__(self, renderer, agent_size):
        self.renderer = renderer
        self.agent_size = agent_size

    def render(self, x, u=None):
        x_dim = x.shape[0] // self.agent_size
        if u is not None:
            u_dim = u.shape[0] // self.agent_size
        
        image = None
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            if u is not None:
                u_agent = u[u_dim*i:u_dim*(i+1)]
            else:
                u_agent = None            
            image = self.renderer.render(x_agent, u_agent, override_image=image)
        return image


class MultiAgentRewardSystem(object):
    """
    RewardSystem wrapper for Multi Agent environment.
    """    
    def __init__(self, reward_system, agent_size):
        self.reward_system = reward_system
        self.agent_size = agent_size

    def reset(self):
        self.reward_system.reset()

    def render(self, image):
        return self.reward_system.render(image)

    def evaluate(self, x, dt):
        x_dim = x.shape[0] // self.agent_size        
        evaluated_reward = 0.0
        for i in range(self.agent_size):
            x_agent = x[x_dim*i:x_dim*(i+1)]
            evaluated_reward += self.reward_system.evaluate(x_agent, dt)
        return evaluated_reward
